from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Any, Callable, Dict, List, Optional, TypeVar

from kiota_abstractions.serialization import ParseNode, SerializationWriter

T = TypeVar('T')


@dataclass
class User2():
    additional_data: Dict[str, Any] = field(default_factory=dict)
    id: Optional[int] = None
    display_name: Optional[str] = None
    age: Optional[int] = None
    gpa: Optional[float] = None

    @staticmethod
    def create_from_discriminator_value(parse_node: Optional[ParseNode] = None) -> User2:
        """
        Creates a new instance of the appropriate class based on discriminator value
        Args:
            parseNode: The parse node to use to read the discriminator value and create the object
        Returns: Attachment
        """
        if not parse_node:
            raise TypeError("parse_node cannot be null")
        return User2()

    def get_field_deserializers(self) -> Dict[str, Callable[[ParseNode], None]]:
        """Gets the deserialization information for this object.

        Returns:
            Dict[str, Callable[[ParseNode], None]]: The deserialization information for this
            object where each entry is a property key with its deserialization callback.
        """
        return {
            "id": lambda n: setattr(self, 'id', n.get_int_value()),
            "display_name": lambda n: setattr(self, 'display_name', n.get_str_value()),
            "age": lambda n: setattr(self, 'age', n.get_int_value()),
            "gpa": lambda n: setattr(self, 'gpa', n.get_float_value())
        }

    def serialize(self, writer: SerializationWriter) -> None:
        """Writes the objects properties to the current writer.

        Args:
            writer (SerializationWriter): The writer to write to.
        """
        if not writer:
            raise TypeError("Writer cannot be null")
        writer.write_int_value("id", self.id)
        writer.write_str_value("display_name", self.display_name)
        writer.write_int_value("age", self.age)
        writer.write_float_value("gpa", self.gpa)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional

from kiota_abstractions.serialization import Parsable, ParseNode, SerializationWriter
from kiota_abstractions.serialization.parse_node_helper import ParseNodeHelper

from . import User, User2


@dataclass
class UnionType(Parsable):
    additional_data: Dict[str, Any] = field(default_factory=dict)
    composed_type1: Optional[User] = None
    composed_type2: Optional[User2] = None
    string_value: Optional[str] = None
    composed_type3: Optional[List[User]] = None

    @staticmethod
    def create_from_discriminator_value(parse_node: Optional[ParseNode] = None) -> UnionType:
        """
        Creates a new instance of the appropriate class based on discriminator value
        Args:
            parseNode: The parse node to use to read the discriminator value and create the object
        Returns: Attachment
        """
        if not parse_node:
            raise TypeError("parse_node cannot be null")

        try:
            mapping_value = parse_node.get_child_node("@odata.type").get_str_value()
        except AttributeError:
            mapping_value = None
        result = UnionType()
        if mapping_value and mapping_value.casefold() == "#microsoft.graph.User".casefold():
            result.composed_type1 = User()
        elif mapping_value and mapping_value.casefold() == "#microsoft.graph.User2".casefold():
            result.composed_type2 = User2()
        elif string_value := parse_node.get_str_value():
            result.string_value = string_value
        elif values := parse_node.get_collection_of_object_values(User):
            result.composed_type3 = values
        return result

    def get_field_deserializers(self) -> Dict[str, Callable[[ParseNode], None]]:
        """Gets the deserialization information for this object.

        Returns:
            Dict[str, Callable[[ParseNode], None]]: The deserialization information for this
            object where each entry is a property key with its deserialization callback.
        """
        if self.composed_type1:
            return self.composed_type1.get_field_deserializers()
        if self.composed_type2:
            return self.composed_type2.get_field_deserializers()
        # composed_type3 is omitted on purpose
        return {}

    def serialize(self, writer: SerializationWriter) -> None:
        """Writes the objects properties to the current writer.

        Args:
            writer (SerializationWriter): The writer to write to.
        """
        if not writer:
            raise TypeError("Writer cannot be null")

        if self.composed_type1:
            writer.write_object_value(None, self.composed_type1)
        elif self.composed_type2:
            writer.write_object_value(None, self.composed_type2)
        elif self.composed_type3:
            writer.write_collection_of_object_values(None, self.composed_type3)
        elif self.string_value:
            writer.write_str_value(None, self.string_value)



import json

import pytest

url: str = "https://graph.microsoft.com/v1.0/$metadata#users/$entity"


@pytest.fixture
def user1_json():
    return '{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
            '"office_location": "dunhill", '\
            '"updated_at": "2021-07-29T03:07:25Z", '\
            '"birthday": "2000-09-04", '\
            '"business_phones": ["+1 205 555 0108"], '\
            '"mobile_phone": null, '\
            '"is_active": true}'


@pytest.fixture
def user2_json():

    return '{"id": 2, "display_name": "MOD Administrator", "age": 32, "gpa": 3.9}'


@pytest.fixture
def users_json():
    return '[{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
            '"office_location": "dunhill", '\
            '"updated_at": "2021-07-29T03:07:25Z", '\
            '"birthday": "2000-09-04", '\
            '"business_phones": ["+1 205 555 0108"], '\
            '"mobile_phone": null, '\
            '"is_active": true}, '\
            '{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
            '"office_location": "dunhill", '\
            '"updated_at": "2021-07-29T03:07:25Z", '\
            '"birthday": "2000-09-04", '\
            '"business_phones": ["+1 205 555 0108"], '\
            '"mobile_phone": null, '\
            '"is_active": true}]'


@pytest.fixture
def sample_entity_json():

    entity_json = json.dumps(
        {
            "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#$entity",
            "id": "8f841f30-e6e3-439a-a812-ebd369559c36"
        }
    )
    return entity_json



import pytest

from kiota_serialization_json.json_serialization_writer import JsonSerializationWriter
from kiota_serialization_json.json_serialization_writer_factory import (
    JsonSerializationWriterFactory,
)


def test_get_serialization_writer():
    factory = JsonSerializationWriterFactory()
    writer = factory.get_serialization_writer('application/json')
    assert isinstance(writer, JsonSerializationWriter)


def test_get_serialization_writer_no_content_type():
    with pytest.raises(TypeError) as e_info:
        factory = JsonSerializationWriterFactory()
        factory.get_serialization_writer('')


def test_get_serialization_writer_unsupported_content_type():
    with pytest.raises(Exception) as e_info:
        factory = JsonSerializationWriterFactory()
        factory.get_serialization_writer('application/xml')


def test_get_valid_content_type():
    factory = JsonSerializationWriterFactory()
    content_type = factory.get_valid_content_type()
    assert content_type == 'application/json'



from uuid import UUID

import pytest
from dateutil import parser

from kiota_serialization_json.json_serialization_writer import JsonSerializationWriter

from ..helpers import OfficeLocation, User, User2


@pytest.fixture
def user_1():
    user = User()
    user.updated_at = parser.parse("2022-01-27T12:59:45.596117")
    user.is_active = True
    user.id = UUID("8f841f30-e6e3-439a-a812-ebd369559c36")
    return user


@pytest.fixture
def user_2():
    user = User2()
    user.age = 32
    user.display_name = "John Doe"
    return user


def test_write_str_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_str_value("displayName", "Adele Vance")
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"displayName": "Adele Vance"}'


def test_write_string_value_no_key():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_str_value(None, "Adele Vance")
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '"Adele Vance"'


def test_write_bool_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_bool_value("isActive", True)
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"isActive": true}'


def test_write_int_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_int_value("timestamp", 28192199291929192)
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"timestamp": 28192199291929192}'


def test_write_float_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_float_value("gpa", 3.2)
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"gpa": 3.2}'


def test_write_uuid_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_uuid_value("id", UUID("8f841f30-e6e3-439a-a812-ebd369559c36"))
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"id": "8f841f30-e6e3-439a-a812-ebd369559c36"}'


def test_write_datetime_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_datetime_value(
        "updatedAt", parser.parse('2022-01-27T12:59:45.596117')
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"updatedAt": "2022-01-27T12:59:45.596117"}'


def test_write_timedelta_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_timedelta_value(
        "diff",
        parser.parse('2022-01-27T12:59:45.596117') - parser.parse('2022-01-27T10:59:45.596117')
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"diff": "2:00:00"}'


def test_write_date_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_date_value("birthday", parser.parse("2000-09-04").date())
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"birthday": "2000-09-04"}'


def test_write_time_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_time_value(
        "time",
        parser.parse('2022-01-27T12:59:45.596117').time()
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"time": "12:59:45.596117"}'


def test_write_collection_of_primitive_values():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_collection_of_primitive_values(
        "businessPhones", ["+1 412 555 0109", 1]
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"businessPhones": ["+1 412 555 0109", 1]}'


def test_write_collection_of_object_values(user_1, user_2):
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_collection_of_object_values("users", [user_1, user_2])
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"users": [{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
        '"updated_at": "2022-01-27T12:59:45.596117", "is_active": true}, '\
        '{"display_name": "John Doe", "age": 32}]}'


def test_write_collection_of_enum_values():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_collection_of_enum_values(
        "officeLocation", [OfficeLocation.Dunhill, OfficeLocation.Oval]
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"officeLocation": ["dunhill", "oval"]}'


def test_write_object_value(user_1):
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_object_value("user1", user_1)
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"user1": {"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
        '"updated_at": "2022-01-27T12:59:45.596117", "is_active": true}}'


def test_write_enum_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_enum_value("officeLocation", OfficeLocation.Dunhill)
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"officeLocation": "dunhill"}'


def test_write_null_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_null_value("mobilePhone")
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"mobilePhone": null}'


def test_write_additional_data_value():
    json_serialization_writer = JsonSerializationWriter()
    json_serialization_writer.write_additional_data_value(
        {
            "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
            "businessPhones": ["+1 205 555 0108"],
        }
    )
    content = json_serialization_writer.get_serialized_content()
    content_string = content.decode('utf-8')
    assert content_string == '{"@odata.context": '\
        '"https://graph.microsoft.com/v1.0/$metadata#users/$entity", '\
            '"businessPhones": ["+1 205 555 0108"]}'



import json
from uuid import UUID

import pytest

from kiota_serialization_json.json_parse_node import JsonParseNode
from kiota_serialization_json.json_serialization_writer import JsonSerializationWriter

from ..helpers import OfficeLocation, User, User2
from ..helpers.union_type import UnionType

display_name = "John Doe"


class TestUnionWrapperParseTests:

    def test_parse_union_type_complex_property1(self):
        json_string = '{"@odata.type": "#microsoft.graph.User",\
            "id": "8f841f30-e6e3-439a-a812-ebd369559c36","office_location": "dunhill",\
                "is_active": true, "updated_at": "2021-07-29T03:07:25Z",\
                    "birthday": "2000-09-04", "business_phones": ["+1 205 555 0108"],\
                        "mobile_phone": null}'

        parse_node = JsonParseNode(json.loads(json_string))
        result = parse_node.get_object_value(UnionType)

        assert result.composed_type1
        assert result.additional_data == {"@odata.type": "#microsoft.graph.User"}
        assert not result.composed_type2
        assert not result.composed_type3
        assert not result.string_value
        assert result.composed_type1.id == UUID("8f841f30-e6e3-439a-a812-ebd369559c36")
        assert result.composed_type1.office_location == OfficeLocation.Dunhill
        assert result.composed_type1.is_active is True
        assert str(result.composed_type1.updated_at) == '2021-07-29 03:07:25+00:00'

    def test_parse_union_type_complex_property2(self):
        json_string = '{"@odata.type": "#microsoft.graph.User2", "id": 43,\
            "display_name": "John Doe", "age": 32, "gpa": 3.9}'

        parse_node = JsonParseNode(json.loads(json_string))
        result = parse_node.get_object_value(UnionType)

        assert not result.composed_type1
        assert result.additional_data == {"@odata.type": "#microsoft.graph.User2"}
        assert result.composed_type2
        assert not result.composed_type3
        assert not result.string_value
        assert result.composed_type2.id == 43
        assert result.composed_type2.display_name == display_name
        assert result.composed_type2.age == 32
        assert result.composed_type2.gpa == 3.9

    def test_parse_union_type_complex_property3(self):
        json_string = '[{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", "display_name": "John Doe",\
            "age": 32, "gpa": 3.9, "office_location": "dunhill"},\
                {"id": "8f841f30-e6e3-439a-a812-ebd369559c36", "display_name": "Jane Doe",\
                    "age": 30, "gpa": 3.8, "office_location": "oval"}]'

        parse_node = JsonParseNode(json.loads(json_string))
        result = parse_node.get_object_value(UnionType)

        assert not result.composed_type1
        assert not result.composed_type2
        assert result.composed_type3
        assert not result.string_value
        assert len(result.composed_type3) == 2
        assert result.composed_type3[1].office_location == OfficeLocation.Oval
        assert result.composed_type3[0].office_location == OfficeLocation.Dunhill

    def test_parse_union_type_string_value(self):
        json_string = json.dumps(display_name)
        parse_node = JsonParseNode(json.loads(json_string))
        result = parse_node.get_object_value(UnionType)

        assert not result.composed_type1
        assert not result.composed_type2
        assert not result.composed_type3
        assert result.string_value
        assert result.string_value == display_name

    def test_serializes_union_type_string_value(self):
        writer = JsonSerializationWriter()
        model = UnionType(string_value=display_name)
        model.serialize(writer)

        content = writer.get_serialized_content()
        content_string = content.decode('utf-8')
        assert content_string == json.dumps(display_name)

    def test_serializes_union_type_composed_type1(self):
        writer = JsonSerializationWriter()
        composed_type1 = User(
            id=UUID("8f841f30-e6e3-439a-a812-ebd369559c36"), office_location=OfficeLocation.Dunhill
        )
        composed_type2 = User2(display_name=display_name, age=32)
        model = UnionType(composed_type1=composed_type1, composed_type2=composed_type2)
        model.serialize(writer)

        content = writer.get_serialized_content()
        content_string = content.decode('utf-8')
        assert content_string == '{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
            '"office_location": "dunhill"}'

    def test_serializes_union_type_complex_property2(self):
        writer = JsonSerializationWriter()
        composed_type2 = User2(id=1, display_name=display_name, age=32)
        model = UnionType(composed_type2=composed_type2)
        model.serialize(writer)

        content = writer.get_serialized_content()
        content_string = content.decode('utf-8')
        assert content_string == '{"id": 1, "display_name": "John Doe", "age": 32}'

    def test_serializes_union_type_complex_property3(self):
        writer = JsonSerializationWriter()
        user_1 = User(
            id=UUID("8f841f30-e6e3-439a-a812-ebd369559c36"),
            office_location=OfficeLocation.Dunhill,
            is_active=True
        )
        user_2 = User(
            id=UUID("7031ea06-d26d-4c9a-a738-ce6fe524f05f"),
            office_location=OfficeLocation.Oval,
            is_active=False
        )
        model = UnionType(composed_type3=[user_1, user_2])
        model.serialize(writer)

        content = writer.get_serialized_content()
        content_string = content.decode('utf-8')
        assert content_string == '[{"id": "8f841f30-e6e3-439a-a812-ebd369559c36", '\
            '"office_location": "dunhill", "is_active": true}, '\
                '{"id": "7031ea06-d26d-4c9a-a738-ce6fe524f05f", "office_location": "oval", '\
                    '"is_active": false}]'



import json
from datetime import date, datetime, time, timedelta
from uuid import UUID

import pytest

from kiota_serialization_json.json_parse_node import JsonParseNode

from ..helpers import OfficeLocation, User

url: str = "https://graph.microsoft.com/v1.0/$metadata#users/$entity"


def test_get_str_value():
    parse_node = JsonParseNode("Diego Siciliani")
    result = parse_node.get_str_value()
    assert result == "Diego Siciliani"


def test_get_int_value():
    parse_node = JsonParseNode(1454)
    result = parse_node.get_int_value()
    assert result == 1454


def test_get_bool_value():
    parse_node = JsonParseNode(False)
    result = parse_node.get_bool_value()
    assert result is False


def test_get_float_value():
    parse_node = JsonParseNode(44.6)
    result = parse_node.get_float_value()
    assert result == 44.6


def test_get_uuid_value():
    parse_node = JsonParseNode("f58411c7-ae78-4d3c-bb0d-3f24d948de41")
    result = parse_node.get_uuid_value()
    assert result == UUID("f58411c7-ae78-4d3c-bb0d-3f24d948de41")


def test_get_datetime_value():
    parse_node = JsonParseNode('2022-01-27T12:59:45.596117')
    result = parse_node.get_datetime_value()
    assert isinstance(result, datetime)


def test_get_date_value():
    parse_node = JsonParseNode('2015-04-20T11:50:51Z')
    result = parse_node.get_date_value()
    assert isinstance(result, date)
    assert str(result) == '2015-04-20'


def test_get_time_value():
    parse_node = JsonParseNode('2022-01-27T12:59:45.596117')
    result = parse_node.get_time_value()
    assert isinstance(result, time)
    assert str(result) == '12:59:45.596117'


def test_get_timedelta_value():
    parse_node = JsonParseNode('2022-01-27T12:59:45.596117')
    result = parse_node.get_timedelta_value()
    assert isinstance(result, timedelta)
    assert str(result) == '12:59:45'


def test_get_collection_of_primitive_values():
    parse_node = JsonParseNode([12.1, 12.2, 12.3, 12.4, 12.5])
    result = parse_node.get_collection_of_primitive_values(float)
    assert result == [12.1, 12.2, 12.3, 12.4, 12.5]


def test_get_collection_of_primitive_values_no_type():
    parse_node = JsonParseNode(["One", "Two", "Three", "Four", "Five"])
    result = parse_node.get_collection_of_primitive_values(None)
    assert result == ["One", "Two", "Three", "Four", "Five"]


def test_get_bytes_value():
    parse_node = JsonParseNode('U2Ftd2VsIGlzIHRoZSBiZXN0')
    result = parse_node.get_bytes_value()
    assert isinstance(result, bytes)


def test_get_collection_of_enum_values():
    parse_node = JsonParseNode(["dunhill", "oval"])
    result = parse_node.get_collection_of_enum_values(OfficeLocation)
    assert isinstance(result, list)
    assert result == [OfficeLocation.Dunhill, OfficeLocation.Oval]


def test_get_enum_value():
    parse_node = JsonParseNode("dunhill")
    result = parse_node.get_enum_value(OfficeLocation)
    assert isinstance(result, OfficeLocation)
    assert result == OfficeLocation.Dunhill


def test_get_object_value(user1_json):
    parse_node = JsonParseNode(json.loads(user1_json))
    result = parse_node.get_object_value(User)
    assert isinstance(result, User)
    assert result.id == UUID("8f841f30-e6e3-439a-a812-ebd369559c36")
    assert result.office_location == OfficeLocation.Dunhill
    assert isinstance(result.updated_at, datetime)
    assert isinstance(result.birthday, date)
    assert result.business_phones == ["+1 205 555 0108"]
    assert result.is_active is True
    assert result.mobile_phone is None


def test_get_collection_of_object_values(users_json):
    parse_node = JsonParseNode(json.loads(users_json))
    result = parse_node.get_collection_of_object_values(User)
    assert isinstance(result[0], User)
    assert result[0].id == UUID("8f841f30-e6e3-439a-a812-ebd369559c36")
    assert result[0].office_location == OfficeLocation.Dunhill
    assert isinstance(result[0].updated_at, datetime)
    assert isinstance(result[0].birthday, date)
    assert result[0].business_phones == ["+1 205 555 0108"]
    assert result[0].is_active is True
    assert result[0].mobile_phone is None



from __future__ import annotations

import base64
import json
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, TypeVar
from uuid import UUID

from kiota_abstractions.serialization import Parsable, SerializationWriter

T = TypeVar("T")
U = TypeVar("U", bound=Parsable)


class JsonSerializationWriter(SerializationWriter):

    PROPERTY_SEPARATOR: str = ','

    def __init__(self) -> None:
        self.writer: Dict = {}
        self.value: Any = None

        self._on_start_object_serialization: Optional[Callable[[Parsable, SerializationWriter],
                                                               None]] = None
        self._on_before_object_serialization: Optional[Callable[[Parsable], None]] = None
        self._on_after_object_serialization: Optional[Callable[[Parsable], None]] = None

    def write_str_value(self, key: Optional[str], value: Optional[str]) -> None:
        """Writes the specified string value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[str]): The string value to be written.
        """
        if isinstance(value, str):
            if key:
                self.writer[key] = value
            else:
                self.value = value

    def write_bool_value(self, key: Optional[str], value: Optional[bool]) -> None:
        """Writes the specified boolean value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[bool]): The boolean value to be written.
        """
        if isinstance(value, bool):
            if key:
                self.writer[key] = value
            else:
                self.value = value

    def write_int_value(self, key: Optional[str], value: Optional[int]) -> None:
        """Writes the specified integer value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[int]): The integer value to be written.
        """
        if isinstance(value, int):
            if key:
                self.writer[key] = value
            else:
                self.value = value

    def write_float_value(self, key: Optional[str], value: Optional[float]) -> None:
        """Writes the specified float value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[float]): The float value to be written.
        """
        if isinstance(value, float):
            if key:
                self.writer[key] = value
            else:
                self.value = value

    def write_uuid_value(self, key: Optional[str], value: Optional[UUID]) -> None:
        """Writes the specified uuid value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[UUId]): The uuid value to be written.
        """
        if isinstance(value, UUID):
            if key:
                self.writer[key] = str(value)
            else:
                self.value = str(value)

    def write_datetime_value(self, key: Optional[str], value: Optional[datetime]) -> None:
        """Writes the specified datetime offset value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[datetime]): The datetime offset value to be written.
        """
        if isinstance(value, datetime):
            if key:
                self.writer[key] = str(value.isoformat())
            else:
                self.value = str(value.isoformat())

    def write_timedelta_value(self, key: Optional[str], value: Optional[timedelta]) -> None:
        """Writes the specified timedelta value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[timedelta]): The timedelta value to be written.
        """
        if isinstance(value, timedelta):
            if key:
                self.writer[key] = str(value)
            else:
                self.value = str(value)

    def write_date_value(self, key: Optional[str], value: Optional[date]) -> None:
        """Writes the specified date value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[date]): The date value to be written.
        """
        if isinstance(value, date):
            if key:
                self.writer[key] = str(value)
            else:
                self.value = str(value)

    def write_time_value(self, key: Optional[str], value: Optional[time]) -> None:
        """Writes the specified time value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[time]): The time value to be written.
        """
        if isinstance(value, time):
            if key:
                self.writer[key] = str(value)
            else:
                self.value = str(value)

    def write_collection_of_primitive_values(
        self, key: Optional[str], values: Optional[List[T]]
    ) -> None:
        """Writes the specified collection of primitive values to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values (Optional[List[T]]): The collection of primitive values to be written.
        """
        if isinstance(values, list):
            result = []
            for val in values:
                temp_writer = self._create_new_writer()
                temp_writer.write_any_value(None, val)
                result.append(temp_writer.value)

            if key:
                self.writer[key] = result
            else:
                self.value = result

    def write_collection_of_object_values(
        self, key: Optional[str], values: Optional[List[U]]
    ) -> None:
        """Writes the specified collection of model objects to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values (Optional[List[U]]): The collection of model objects to be written.
        """
        if isinstance(values, list):
            obj_list = []
            for val in values:
                temp_writer = self._create_new_writer()
                temp_writer.write_object_value(None, val)
                obj_list.append(temp_writer.value)

            if key:
                self.writer[key] = obj_list
            else:
                self.value = obj_list

    def write_collection_of_enum_values(
        self, key: Optional[str], values: Optional[List[Enum]]
    ) -> None:
        """Writes the specified collection of enum values to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            values Optional[List[Enum]): The enum values to be written.
        """
        if isinstance(values, list):
            result = []
            for val in values:
                temp_writer = self._create_new_writer()
                temp_writer.write_enum_value(None, val)
                result.append(temp_writer.value)

            if key:
                self.writer[key] = result
            else:
                self.value = result

    def write_bytes_value(self, key: Optional[str], value: bytes) -> None:
        """Writes the specified byte array as a base64 string to the stream with an optional
        given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (bytes): The byte array to be written.
        """
        if isinstance(value, bytes):
            base64_bytes = base64.b64encode(value)
            base64_string = base64_bytes.decode('utf-8')
            if key:
                self.writer[key] = base64_string
            else:
                self.value = base64_string

    def write_object_value(
        self, key: Optional[str], value: Optional[U], *additional_values_to_merge: U
    ) -> None:
        """Writes the specified model object to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Parsable): The model object to be written.
            additional_values_to_merge (tuple[Parsable]): The additional values to merge to the
            main value when serializing an intersection wrapper.
        """
        if value or additional_values_to_merge:
            temp_writer = self._create_new_writer()

            if value:
                self._serialize_value(temp_writer, value)

            if additional_values_to_merge:
                for additional_value in filter(lambda x: x is not None, additional_values_to_merge):
                    self._serialize_value(temp_writer, additional_value)
                    if on_after := self.on_after_object_serialization:
                        on_after(additional_value)

            if value and self._on_after_object_serialization:
                self._on_after_object_serialization(value)

            if key:
                self.writer[key] = temp_writer.writer
            else:
                self.value = temp_writer.writer

    def write_enum_value(self, key: Optional[str], value: Optional[Enum]) -> None:
        """Writes the specified enum value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (Optional[Enum]): The enum value to be written.
        """
        if isinstance(value, Enum):
            if key:
                self.writer[key] = value.value
            else:
                self.value = value.value

    def write_null_value(self, key: Optional[str]) -> None:
        """Writes a null value for the specified key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
        """
        if key:
            self.writer[key] = None
        else:
            self.value = "null"

    def write_additional_data_value(self, value: Dict[str, Any]) -> None:
        """Writes the specified additional data to the stream.
        Args:
            value (Dict[str, Any]): he additional data to be written.
        """
        if isinstance(value, dict):
            for key, val in value.items():
                self.writer[key] = val

    def get_serialized_content(self) -> bytes:
        """Gets the value of the serialized content.
        Returns:
            bytes: The value of the serialized content.
        """
        if self.writer and self.value:
            # Json output is invalid if it has a mix of values
            # and key-value pairs.
            raise ValueError("Invalid Json output")

        if self.value:
            json_string = json.dumps(self.value)
            self.value = None
        else:
            json_string = json.dumps(self.writer)
            self.writer.clear()

        stream = json_string.encode('utf-8')
        return stream

    @property
    def on_before_object_serialization(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called before the object gets serialized.
        Returns:
            Optional[Callable[[Parsable], None]]:the callback called before the object
            gets serialized.
        """
        return self._on_before_object_serialization

    @on_before_object_serialization.setter
    def on_before_object_serialization(self, value: Optional[Callable[[Parsable], None]]) -> None:
        """Sets the callback called before the objects gets serialized.
        Args:
            value (Optional[Callable[[Parsable], None]]): the callback called before the objects
            gets serialized.
        """
        self._on_before_object_serialization = value

    @property
    def on_after_object_serialization(self) -> Optional[Callable[[Parsable], None]]:
        """Gets the callback called after the object gets serialized.
        Returns:
            Optional[Optional[Callable[[Parsable], None]]]: the callback called after the object
            gets serialized.
        """
        return self._on_after_object_serialization

    @on_after_object_serialization.setter
    def on_after_object_serialization(self, value: Optional[Callable[[Parsable], None]]) -> None:
        """Sets the callback called after the objects gets serialized.
        Args:
            value (Optional[Callable[[Parsable], None]]): the callback called after the objects
            gets serialized.
        """
        self._on_after_object_serialization = value

    @property
    def on_start_object_serialization(
        self
    ) -> Optional[Callable[[Parsable, SerializationWriter], None]]:
        """Gets the callback called right after the serialization process starts.
        Returns:
            Optional[Callable[[Parsable, SerializationWriter], None]]: the callback called
            right after the serialization process starts.
        """
        return self._on_start_object_serialization

    @on_start_object_serialization.setter
    def on_start_object_serialization(
        self, value: Optional[Callable[[Parsable, SerializationWriter], None]]
    ) -> None:
        """Sets the callback called right after the serialization process starts.
        Args:
            value (Optional[Callable[[Parsable, SerializationWriter], None]]): the callback
            called right after the serialization process starts.
        """
        self._on_start_object_serialization = value

    def write_non_parsable_object_value(self, key: Optional[str], value: T) -> None:
        """Writes the specified value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value (object): The value to be written.
        """
        if hasattr(value, '__dict__'):
            if key:
                self.writer[key] = value.__dict__
            else:
                self.value = value.__dict__

    def write_any_value(self, key: Optional[str], value: Any) -> Any:
        """Writes the specified value to the stream with an optional given key.
        Args:
            key (Optional[str]): The key to be used for the written value. May be null.
            value Any): The value to be written.
        """
        primitive_types = [bool, str, int, float, UUID, datetime, timedelta, date, time, Enum]
        if value:
            value_type = type(value)
            if key:
                if value_type in primitive_types:
                    method = getattr(self, f'write_{value_type.__name__.lower()}_value')
                    method(key, value)
                elif isinstance(value, Parsable):
                    self.write_object_value(key, value)
                elif hasattr(value, '__dict__'):
                    self.write_non_parsable_object_value(key, value)
                else:
                    raise TypeError(
                        f"Encountered an unknown type during serialization {value_type} \
                            with key {key}"
                    )
            else:
                if value_type in primitive_types:
                    method = getattr(self, f'write_{value_type.__name__.lower()}_value')
                    method(None, value)
                elif isinstance(value, Parsable):
                    self.write_object_value(None, value)
                elif hasattr(value, '__dict__'):
                    self.write_non_parsable_object_value(None, value)
                else:
                    raise TypeError(
                        f"Encountered an unknown type during serialization {value_type}"
                    )

    def _serialize_value(self, temp_writer: JsonSerializationWriter, value: U):
        if on_before := self.on_before_object_serialization:
            on_before(value)
        if on_start := self.on_start_object_serialization:
            on_start(value, self)
        value.serialize(temp_writer)

    def _create_new_writer(self) -> SerializationWriter:
        writer = JsonSerializationWriter()
        writer.on_before_object_serialization = self.on_before_object_serialization
        writer.on_after_object_serialization = self.on_after_object_serialization
        writer.on_start_object_serialization = self.on_start_object_serialization
        return writer
